from parser_base import *
from mel_ast import *


class MELError(Exception):
    """Класс для исключений интерпретатора"""

    def __init__(self, message):
        super().__init__(message)


class MELParser(BaseParser):
    """Простейший интерпретатор MEL (Math Expr Lang).
    Вычисления производятся в процессе разбора

    Грамматика:
        num     -> <число>
        ident   -> <идентификатор>
        group   -> num | '(' expr ')'
        mult    -> group (('*' | '/') group)*
        add     -> mult (('+' | '-') mult)*
        expr    -> add
        start   -> expr

    Грамматика 2:
        num     -> <число>
        ident   -> <идентификатор>
        group   -> num | ident | '(' expr ')'
        mult    -> group (('*' | '/') group)*
        add     -> mult (('+' | '-') mult)*
        expr    -> add
        stmt    -> 'read' ident | 'print' expr | ident '=' expr
        prog    -> stmt*
        start   -> prog

    Грамматика 3:
        num     -> <число>
        ident   -> <идентификатор>
        group   -> num | ident | func_call | '(' expr ')'
        mult    -> group (('*' | '/') group)*
        add     -> mult (('+' | '-') mult)*
        expr    -> add
        param   -> expr
        params  -> (param (',' param)*)?
        func_call   -> ident ('(' params ')')?
        if      -> 'if' '(' expr ')' stmt ( else stmt )?
        while   -> 'while' '(' expr ')' stmt
        for     -> 'for' '(' stmt? ';' expr? ';' stmt? ')' stmt
        stmt    ->
            '{' stmts '}' |
            if |
            while |
            for |
            'read' ident |
            'print' expr |
            ident '=' expr |
            func_call
        stmts   -> stmt*
        prog    -> stmts
        start   -> prog
    """

    def __init__(self, text: str):
        """Конструктор
        :param text: текст программы
        """
        super().__init__(text)

    def num(self) -> NumNode:
        """num   -> <число>"""
        self.ws()
        pos = self.pos
        while self.curr.isdecimal() or self.curr == '.':
            self.pos += 1
        res = NumNode(float(self.text[pos:self.pos]))
        self.ws()
        return res

    def ident(self) -> IdentNode:
        """ ident   -> <идентификатор> """
        self.ws()
        pos = self.pos
        while self.curr.isalnum():
            self.pos += 1
        res = IdentNode(self.text[pos:self.pos])
        self.ws()
        return res

    def group(self) -> ExprNode:
        """ group   -> num | ident | func_call | '(' expr ')' """
        if self.is_parse('('):
            self.parse('(')
            res = self.expr()
            self.parse(')')
            return res
        elif self.curr.isdecimal() or self.curr == '.':
            return self.num()
        else:
            pos = self.pos
            res = self.ident()
            if self.is_parse('('):
                self.pos = pos
                res = self.func_call()
            return res

    def mult(self) -> BinOpNode:
        """ mult    -> group (('*' | '/') group)* """
        res = self.group()
        while self.is_parse('*', '/'):
            op = self.parse('*', '/')
            tmp = self.group()
            res = BinOpNode(BinOp(op), res, tmp)
        return res

    def add(self) -> BinOpNode:
        """ add     -> mult (('+' | '-') mult)* """
        res = self.mult()
        while self.is_parse('+', '-'):
            op = self.parse('+', '-')
            tmp = self.mult()
            res = BinOpNode(BinOp(op), res, tmp)
        return res

    def expr(self) -> ExprNode:
        """ expr    -> add """
        return self.add()

    def param(self) -> ExprNode:
        """param  -> add"""
        return self.add()

    def params(self) -> Tuple[ExprNode]:
        """ params  -> (param(',' param) *)? """
        params = []
        if not self.is_parse(')'):
            params.append(self.param())
            while self.is_parse(','):
                self.parse(',')
                params.append(self.param())
        return *params,

    def func_call(self) -> FuncCall:
        """ func_call   -> ident ('(' params ')')? """
        ident = self.ident()
        self.parse('(')
        res = FuncCall(ident, *self.params())
        self.parse(')')
        return res

    def if_(self) -> IfNode:
        """ if  -> 'if' '(' expr ')' stmt ( else stmt )? """
        self.parse('if')
        self.parse('(')
        cond = self.expr()
        self.parse(')')
        then_stmt = self.stmt()
        else_stmt = None
        if self.is_parse('else'):
            self.parse('else')
            else_stmt = self.stmt()
        return IfNode(cond, then_stmt, else_stmt)

    def while_(self) -> WhileNode:
        """ while   -> 'while' '(' expr ')' stmt """
        self.parse('while')
        self.parse('(')
        cond = self.expr()
        self.parse(')')
        body = self.stmt()
        return WhileNode(cond, body)

    def for_(self) -> ForNode:
        """ for     -> 'for' '(' stmt? ';' expr? ';' stmt? ')' stmt """
        self.parse('for')
        self.parse('(')
        init = None
        if not self.is_parse(';'):
            init = self.stmt()
        self.parse(';')
        cond = None
        if not self.is_parse(';'):
            cond = self.expr()
        self.parse(';')
        step = None
        if not self.is_parse(')'):
            step = self.stmt()
        self.parse(')')
        body = self.stmt()
        return ForNode(init, cond, step, body)

    def stmt(self) -> StmtNode:
        """
        stmt   ->
            '{' stmts '}' |
            if |
            while |
            for |
            'read' ident |
            'print' expr |
            ident '=' expr |
            func_call
        """
        if self.is_parse('{'):
            self.parse('{')
            res = self.stmts()
            self.parse('}')
            return res
        if self.is_parse('if'):
            return self.if_()
        if self.is_parse('while'):
            return self.while_()
        if self.is_parse('for'):
            return self.for_()
        if self.is_parse('read', 'print'):
            op = self.parse('read', 'print')
            if op == 'read':
                return ReadNode(self.ident())
            else:
                return PrintNode(self.expr())
        else:
            pos = self.pos
            ident = self.ident()
            if self.is_parse('('):
                self.pos = pos
                return self.func_call()
            else:
                self.parse('=')
                return AssignNode(ident, self.expr())

    def stmts(self) -> StmtListNode:
        """ stmts   -> stmt* """
        stmts = []
        self.ws()
        while self.curr not in ('$', '}'):
            stmts.append(self.stmt())
        return StmtListNode(*stmts)

    def prog(self) -> StmtListNode:
        """ prog    -> stmts """
        return self.stmts()

    def start(self) -> StmtListNode:
        """ start   -> prog """
        return self.prog()

    @classmethod
    def parse_prog(cls, prog: str) -> StmtListNode:
        parser = cls(str(prog))
        return parser.start()
